% This is the implementation of the algorithm provided in 
%   Bibliographical Connections for Semiparametric Analysis 
%   in Case-Control Studies on Gene-Environment Interactions.
% To reproduce the simulation result presented by the paper, 
%   simply put this script inside Matlab and click on Run button.
% The result will be output to a file named 
%   "Semiparametric_Profile_Likelihood_Algorithm_Implementation_Output.xlsx"
%   under the same directory of wherever this script resides in.
% To use this implementation for your own simulation scenarios,
%   modify some key variables (such as p, n0, n1, etc.) and 
%   the generateData() function will be necessary.

function Semiparametric_Profile_Likelihood_Algorithm_Implementation()
    setenv('BLAS_VERSION','')

    % Variable definitions
    currentRound = 1;
    totalRound = 60;
    internalRound = 6;
    pi1_output_true = zeros(totalRound,1);
    pi1_output_init = zeros(totalRound,1);
    pi1_output_mid = zeros(internalRound*totalRound,1);
    pi1_output_fin = zeros(totalRound,1);
    theta_original = zeros(totalRound,4);
    theta_output = zeros(totalRound,4);
    variance_LogisticRegression = zeros(totalRound, 4);
    variance_SuperSolver = zeros(totalRound, 4);

    % Simulation starts
    while (currentRound <= totalRound)
        % Step 1: Set initial values
        n0 = 100;
        n1 = 100;
        n = n0 + n1;
        p = 0.6;
        [D, X, Z, pi_true] = generateData(n, n0, n1, -3.61, [0.76, 0.36, -0.63], p);
        pi1_output_true(currentRound) = pi_true;
        theta_tilde = initialTheta(pi_true,n0,n1,D,X,Z);

        gamma = getGamma(n,X);
        gamma = densityScale(gamma);
        xi = Z/sum(Z);
        xi = densityNormalize(xi);
        
        % Step 2: Calculate
        m = @(x,z,theta_tilde) theta_tilde(2)*x + theta_tilde(3)*z + theta_tilde(4)*x*z;
        hh = @(x,z,theta_tilde) theta_tilde(1) + m(x,z,theta_tilde);
        h = @(d,x,z,theta_tilde) exp(d * (hh(x,z,theta_tilde))) / (1 + exp(hh(x,z,theta_tilde)));
        
        % Step 7: Repeat until theta_hat converged
        theta_temp = zeros(4,1);
        for time = 1:internalRound
            % Step 5: Update pi_1 and pi_0
            pi_1 = pi1(n, gamma, xi, theta_tilde, X, Z, h); % 0 < pi_1 < 1
            if (pi_1 >= 1 || pi_1 <= 0)
                break;
            end
            pi_0 = 1 - pi_1;
            if (time == 1)
                pi1_output_init(currentRound) = pi_1;
            else
                pi1_output_mid(time+3*(currentRound-1)) = pi_1;
            end
            
            % Step 3 & Step 4: Update the density estimations
            [gamma, xi] = density(n,n0,n1,pi_0,pi_1,gamma,xi,theta_tilde,X,Z,h);
            gamma = densityScale(gamma);
            xi = densityNormalize(xi);
            
            % Step 6: Update theta
            ff = [-3.73784658141544; 0.64165556503607; 0.331832048835377; -0.578882848131804];
            options = optimoptions(@fsolve, 'Algorithm', 'levenberg-marquardt', 'OutputFcn', @custom_stop_fun, 'MaxFunEval', 100);
            theta_hat = fsolve(@(theta_var) equation(theta_var,n,D,X,Z,h,gamma,xi), ff, options);
            disp(theta_hat)
            if max(abs(theta_temp - theta_hat)) < 0.05
                break;
            else
                theta_temp = theta_hat;
            end
        end
        pi1_output_fin(currentRound) = pi1(n, gamma, xi, theta_hat, X, Z, h);
        theta_original(currentRound,:) = theta_tilde';
        variance_LogisticRegression(currentRound,:) = calculateLogisticRegressionVariance(theta_tilde,n,D,X,Z,h,gamma,xi);
        theta_output(currentRound,:) = theta_hat';
        variance_SuperSolver(currentRound,:) = calculateSuperSolverVariance(theta_hat,n,D,X,Z,h,gamma,xi);
        currentRound = currentRound + 1;
    end
    
    % Output results
    pi1_output_mid = pi1_output_mid(pi1_output_mid~=0);
    filename = './Semiparametric_Profile_Likelihood_Algorithm_Implementation_Output.xlsx';
    T = table(pi1_output_true);
    writetable(T, filename, 'Sheet', 1);
    T = table(pi1_output_init);
    writetable(T, filename, 'Sheet', 2);
    T = table(pi1_output_mid);
    writetable(T, filename, 'Sheet', 3);
    T = table(pi1_output_fin);
    writetable(T, filename, 'Sheet', 4);
    T = table(theta_original);
    writetable(T, filename, 'Sheet', 5);
    T = table(theta_output);
    writetable(T, filename, 'Sheet', 6);
    T = table(variance_LogisticRegression);
    writetable(T, filename, 'Sheet', 7);
    T = table(variance_SuperSolver);
    writetable(T, filename, 'Sheet', 8);
end

function theta_tilde = initialTheta(pi1_true,n0,n1,D,X,Z)
    pi0_true = 1 - pi1_true;
    XX = [X,Z];
    model = fitglm(XX, D, 'y ~ x1 * x2', 'Distribution', 'binomial');
    theta_tilde = model.Coefficients.Estimate;
    temp = theta_tilde(1);
    theta_tilde(1) = temp - log(n0 / n1) + log(pi1_true / pi0_true);
end

function [D, X, Z, pi_true] = generateData(n,n0,n1,alpha,beta,p)
    n00 = 0;
    n10 = 0;
    nn = 0;
    i = 1;
    D = zeros(n, 1);
    X = zeros(n, 1);
    Z = zeros(n, 1);
    while ((n10 < n1) || (n00 < n0))
        X(i) = binornd(1,p,1,1);
        Z(i) = normrnd(0,1,1,1);
        nn = nn + 1;
        m = @(beta,x,z) beta(1)*x + beta(2)*z + beta(3)*x*z;
        t = 1 - 1 / (1 + exp(alpha + m(beta,X(i),Z(i))));
        u = rand(1);
        if (u < t)
            D(i) = 1;
            if (n10 < n1)
                n10 = n10 + 1;
                i = i + 1;
            end
        else
            D(i) = 0;
            if (n00 < n0)
                n00 = n00 + 1;
                i = i + 1;
            end
        end
    end
    All = sortrows([D, X, Z], 1);
    D = All(:, 1);
    X = All(:, 2);
    Z = All(:, 3);
    pi_true = n10 / nn;
end

function gamma = getGamma(n,X)
    gamma = ones(n,1);
    a = 1/sum(X(:) == 0);
    b = 1/sum(X(:) == 1);
    for i = 1:n
        if (X(i) == 0)
            gamma(i) = a;
        else
            gamma(i) = b;
        end
    end
end

function scale = densityScale(density)
    scale = density / sum(density);
end

function newD = densityNormalize(raw)
    N = mat2gray(raw);
    newD = N/sum(N);
end

function pi1 = pi1(n,gamma,xi,theta_tilde,X,Z,h)
    Value = 0;
    temp = 0;
    for k = 1:n
        for j = 1:n
            temp = temp + gamma(j) * h(1, X(j), Z(k), theta_tilde);
        end
        Value = Value + temp * xi(k);
        temp = 0;
    end
    pi1 = Value;
end

function [gamma_hat, xi_hat] = density(n,n0,n1,pi_0,pi_1,gamma_tilde,xi_tilde,theta_tilde,X,Z,h)
    gamma_hat = zeros(n, 1);
    xi_hat = zeros(n, 1);
    addend = n1/pi_1;
    coefficient = n0/pi_0 - n1/pi_1;
    sum_xi = 0;
    sum_gamma = 0;
    for a = 1:n
        % a is used as l for gamma_hat and s for xi_hat
        for b = 1:n
            % b is used as j for gamma_tilde and k for xi_tilde
            sum_xi = sum_xi + xi_tilde(b) * h(0,X(a),Z(b),theta_tilde);
            sum_gamma = sum_gamma + gamma_tilde(b) * h(0,X(b),Z(a),theta_tilde);
        end
        gamma_hat(a) = 1/(addend + coefficient * sum_xi);
        xi_hat(a) = 1/(addend + coefficient * sum_gamma);
        sum_xi = 0;
        sum_gamma = 0;
    end
end

function derivative = derivative(theta,d,x,z)
    temp = theta(1) + theta(2)*x + theta(3)*z + theta(4)*x*z;
    temp1 = d * exp(temp) - exp(temp) + d;
    derivative = zeros(1,4);
    derivative(1) = temp1*exp(d*temp)/((1+exp(temp))^2);
    derivative(2) = temp1*exp(d*temp)*x/((1+exp(temp))^2);
    derivative(3) = temp1*exp(d*temp)*z/((1+exp(temp))^2);
    derivative(4) = temp1*exp(d*temp)*x*z/((1+exp(temp))^2);
end

function equation = equation(theta,n,D,X,Z,h,gamma,xi)
    fprintf('theta for this round: %5.4f; %5.4f; %5.4f; %5.4f\n', theta);
    out = zeros(n,4);
    for i = 1:n
        out(i,:) = derivative(theta,D(i),X(i),Z(i)) ./ h(D(i),X(i),Z(i),theta) - equation2(theta,n,D(i),X,Z,gamma,xi,h);
    end
    Equation = zeros(4,1);
    for ii = 1:4
        Equation(ii) = sum(out(:,ii))/n;
    end
    equation = 1000*sum(power(Equation, 2));
end

function equation2 = equation2(theta,n,d,X,Z,gamma,xi,h)
    temp1 = zeros(n,4);
    temp2 = zeros(n,1);
    temp3 = zeros(n,4);
    temp4 = zeros(n,1);
    for j = 1:n
        for k = 1:n
            temp1(k,:) = xi(k) * derivative(theta,d,X(j),Z(k));
            temp2(k) = xi(k) * h(d,X(j),Z(k),theta);
        end
        for jj = 1:4
            temp3(j,jj) = sum(temp1(:,jj)) * gamma(j);
        end
        temp4(j) = sum(temp2) * gamma(j);
    end
    equation2 = zeros(1,4);
    tmp = sum(temp4);
    for m = 1:4
        equation2(m) = sum(temp3(:,m)) ./ tmp;
    end
end

function stop = custom_stop_fun(~, optimValues, ~)
    if optimValues.fval < 2
        stop = true;
    else
        stop = false;
    end
end

function variance = calculateLogisticRegressionVariance(theta,n,D,X,Z,h,gamma,xi)
    numericalDerivative = calculateNumericalDerivative(theta,n,D,X,Z,h,gamma,xi);
    covariance = calculateCovariance(theta,n,D,X,Z,h,gamma,xi);
    numericalDerivative_inverse = pinv(numericalDerivative);
    variance_matrix = numericalDerivative_inverse * covariance * transpose(numericalDerivative_inverse);
    variance = zeros(1,4);
    for i = 1:4
        variance(i) = variance_matrix(i,i);
    end
end

function variance = calculateSuperSolverVariance(theta,n,D,X,Z,h,gamma,xi)
    numericalDerivative = calculateNumericalDerivative(theta,n,D,X,Z,h,gamma,xi);
    covariance = calculateCovariance_bootstrap(theta,n,D,X,Z,gamma,xi,h);
    numericalDerivative_inverse = pinv(numericalDerivative);
    variance_matrix = numericalDerivative_inverse * covariance * transpose(numericalDerivative_inverse);
    variance = zeros(1,4);
    for i = 1:4
        variance(i) = variance_matrix(i,i);
    end
end

function numericalDerivative = calculateNumericalDerivative(theta,n,D,X,Z,h,gamma,xi)
    delta = 0.001;
    theta_left = theta * (1-delta);
    calculation_left = calculateWithThetaHat(theta_left,n,D,X,Z,h,gamma,xi);
    theta_right = theta * (1+delta);
    calculation_right = calculateWithThetaHat(theta_right,n,D,X,Z,h,gamma,xi);
    calculation = calculation_right - calculation_left;
    denominator = 2*delta*theta;
    numericalDerivative = zeros(4,4);
    for i = 1:4
        numericalDerivative(:,i) = calculation / denominator(i);
    end
end

function covariance = calculateCovariance(theta,n,D,X,Z,h,gamma,xi)
    out = zeros(n,4);
    for i = 1:n
        out(i,:) = derivative(theta,D(i),X(i),Z(i)) ./ h(D(i),X(i),Z(i),theta) - equation2(theta,n,D(i),X,Z,gamma,xi,h);
    end
    out_transpose = transpose(out);
    covariance = out_transpose * out / n;
end

function [D_new, X_new, Z_new] = bootstrap_resample(n,D,X,Z)
    D_new = zeros(n,1);
    X_new = zeros(n,1);
    Z_new = zeros(n,1);
    n0 = n/2;
    n1 = n/2;
    for i = 1:n0
        temp = randi(n0);
        D_new(i) = D(temp);
        X_new(i) = X(temp);
        Z_new(i) = Z(temp);
    end
    for i = 1:n1
        temp = randi(n1);
        D_new(n0+i) = D(temp+n0);
        X_new(n0+i) = X(temp+n0);
        Z_new(n0+i) = Z(temp+n0);
    end
end

function covariance = calculateCovariance_bootstrap(theta,n,D,X,Z,gamma,xi,h)
    B = 100;
    out_B = zeros(n*B,4);
    for j = 1:B
        [D_new,X_new,Z_new] = bootstrap_resample(n,D,X,Z);
        for i = 1:n
            out_B(i*B-B+1,:) = derivative(theta,D_new(i),X_new(i),Z_new(i)) ./ h(D_new(i),X_new(i),Z_new(i),theta) - equation2(theta,n,D_new(i),X_new,Z_new,gamma,xi,h);
        end
    end
    out_sum = zeros(n,4);
    for i = 1:B
        out_sum = out_sum + out_B(i*n-n+1:i*n,:);
    end
    out_mean = out_sum / n;
    covariance = zeros(4,4);
    for i = 1:B
        temp = out_B(i*n-n+1:i*n,:) - out_mean(:,:);
        temp_transpose = transpose(temp);
        covariance = covariance + temp_transpose * temp;
    end
    covariance = covariance / (B-1);
end

function calculation = calculateWithThetaHat(theta,n,D,X,Z,h,gamma,xi)
    out = zeros(n,4);
    for i = 1:n
        out(i,:) = derivative(theta,D(i),X(i),Z(i)) ./ h(D(i),X(i),Z(i),theta) - equation2(theta,n,D(i),X,Z,gamma,xi,h);
    end
    calculation = zeros(4,1);
    for ii = 1:4
        calculation(ii) = sum(out(:,ii))/n;
    end
end